#include "config.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <pwd.h>
#include <errno.h>
#include <string.h>

#if HAVE_GETOPT_H
#include <getopt.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif




#include <syslog.h>
#include <signal.h>

#include "misc.h"
#include "get_uid.h"


struct ip_address la, ra;
uint32 lp, rp;
char flagverbose=0, flagsyslog=1, flagprivate=0, flagrandom=0, flaguser=0;

void log_error(struct conn_info c, char *msg, int sys) {
char local[20], remote[20];

	sprintf(local,"%d.%d.%d.%d",c.laddr.d[0],c.laddr.d[1],c.laddr.d[2],c.laddr.d[3]);
	sprintf(remote,"%d.%d.%d.%d",c.raddr.d[0],c.raddr.d[1],c.raddr.d[2],c.raddr.d[3]);

	if(sys) {
		if(flagsyslog) 
	 	 syslog(LOG_DAEMON|LOG_ERR,"error %s:%d %s:%d %s: %s\n",local,c.lport,remote,c.rport,msg,strerror(errno));
		else
		 fprintf(stderr,"error %s:%d %s:%d %s: %s\n",local,c.lport,remote,c.rport,msg,strerror(errno));
	} else {
		if(flagsyslog)
	 	 syslog(LOG_DAEMON|LOG_ERR,"error %s:%d %s:%d %s\n",local,c.lport,remote,c.rport,msg);
		else
		 fprintf(stderr,"error %s:%d %s:%d %s\n",local,c.lport,remote,c.rport,msg);
	}
}

void log_success(struct conn_info c, char *real, char *returned) {
char local[20], remote[20];

	sprintf(local,"%d.%d.%d.%d",c.laddr.d[0],c.laddr.d[1],c.laddr.d[2],c.laddr.d[3]);
	sprintf(remote,"%d.%d.%d.%d",c.raddr.d[0],c.raddr.d[1],c.raddr.d[2],c.raddr.d[3]);
	if(flagsyslog)
		syslog(LOG_DAEMON|LOG_INFO,"ok %s:%d %s:%d \"%s\" \"%s\"",local,c.lport,remote,c.rport,real,returned);
	else {
		fprintf(stderr,"ok %s:%d %s:%d \"%s\" \"%s\"\n",local,c.lport,remote,c.rport,real,returned);
		fflush(stderr);
	}
}
		




void ident_reply(struct conn_info c, char *reply) {
	printf("%d, %d : USERID : UNIX :%s\r\n",c.lport,c.rport,reply);
}

void ident_error(struct conn_info c, char *error) {
	printf("%d, %d : ERROR : %s\r\n",c.lport,c.rport,error);
}


void get_addrs(struct conn_info *c) {
struct sockaddr_in local, remote;
int len;

	len=sizeof(local);
	if(getsockname(0,(struct sockaddr *)&local, &len)==-1) {
		log_error(*c, "could not get local address", 1);
		exit(1);
	}
	memcpy(c->laddr.d,&local.sin_addr,sizeof(c->laddr.d));
	len=sizeof(remote);
	if(getpeername(0,(struct sockaddr *)&remote, &len)==-1) {
		log_error(*c, "could not get remote address", 1);
		exit(1);
	}
	memcpy(c->raddr.d,&remote.sin_addr,sizeof(c->raddr.d));
}

void get_ports(struct conn_info *c) {
char *x, line[512], *lport, *rport;
int len, fields[2];
uint32 l, r;

	x=fgets(line,sizeof(line),stdin);
	if(!x) {
		log_error(*c,"could not read from network",1);
		exit(1);
	}
	if(line[strlen(line)-1]!='\n') {
		log_error(*c,"input too long",0);
		exit(1);
	}
	len=strlen(line);
	tokenize(line,fields,2);
	if((fields[1]>=len)) {
		log_error(*c,"input too short",0);
		exit(1);
	}
	lport=&line[fields[0]]; rport=&line[fields[1]];
	if((lport[scan_uint32(lport,&l)])||(l>65535)) {
		ident_error(*c,"INVALID-PORT");
		log_error(*c,"could not parse local-port",0); 
		exit(1);
	}
	c->lport=l;
	if((rport[scan_uint32(rport,&r)])||(r>65535)) {
		ident_error(*c,"INVALID-PORT");
		log_error(*c,"could not parse remote-port",0);
		exit(1);
	}
	c->rport=r;
}

void timeout(int s) {
	exit(1);
}

void do_help() {
	fprintf(stderr,"Usage: hidentd [-hsv] [-u username] [-p username]\n");
	fprintf(stderr,"\n\t-h\t\thelp\n");
	fprintf(stderr,"\t-v\t\tverbose\n");
	fprintf(stderr,"\t-s\t\tlog to stderr\n");
	fprintf(stderr,"\t-r\t\trespond with random strings\n");
	fprintf(stderr,"\t-o\t\tallow users to specify ident response in ~/.ident\n");
	fprintf(stderr,"\t-u username\trespond with 'username' for masqueraded connections\n");
	fprintf(stderr,"\t-p username\talways respond with 'username' (overrides -u)\n");
	fprintf(stderr,"\nSend bug reports to %s\n", PACKAGE_BUGREPORT);
	exit(0);
}

void get_random(char **username) {
static char name[9];
unsigned char c;
int i;

	name[8]='\0';
	for(i=0;i<4;i++) {
		c=(int)(255.0*rand()/(RAND_MAX+1.0));
		hexchar(c, &name[2*i]);
	}
	*username=name;
}

void get_ownname(char *dir, char **username) {
static char name[32];
struct stat sb;
int fd, len, i;

	if(chdir(dir)==-1) return;
	if(stat(".ident", &sb)==-1) return;
	if(!S_ISREG(sb.st_mode)) return;
	fd=open(".ident", O_RDONLY);
	if(fd==-1) return;
	memset(name, 0, sizeof(name));
	len=read(fd, name, sizeof(name));
	if(len<0) {
		close(fd);
		return;
	}
	name[len-1]='\0';
	for(i=0; i<sizeof(name); i++)
		if((name[i]=='\0')||(name[i]=='\r')||(name[i]=='\n')) {
			name[i]='\0';
			break;
		}
	close(fd);	
	*username=name;
}

struct passwd *get_pw(struct conn_info ci) {
int uid;
	uid=get_uid(ci);
	if(uid<0) return NULL;
	return getpwuid(uid);
}

int main(int argc, char **argv) {
	char *username=NULL, *allname=NULL, *realname=NULL, *defname=NULL;
	int uid=-1, opt;
	struct passwd *pw;
	struct conn_info ci;



	while((opt=getopt(argc, argv, "hvrop:u:s"))!=-1) {
	 switch(opt) {
		 case 'h': do_help(); break;
		 case 'v': flagverbose=1; break;
		 case 'u': defname=optarg; break;
		 case 's': flagsyslog=0;break;
		 case 'p': allname=optarg; break;
		 case 'r': flagrandom=1; srand(time(NULL)); break;
		 case 'o': flaguser=1; break;
		 default:  do_help();
	 }
	}
	memset(&ci,0,sizeof(ci));

#ifndef OS_NEEDS_ROOT
	if(getuid()==0) {
		log_error(ci, "refusing to run as root. Read the documentation.", 0);
		exit(1);
	}
#else
	if(getuid()!=0) {
		log_error(ci, "root privileges are required on this operating system. Read the documentation.", 0);
		exit(1);
	}
#endif

	get_addrs(&ci);
	alarm(60);
	signal(SIGALRM,timeout);
	get_ports(&ci);


	realname="UNKNOWN";
	pw=get_pw(ci);

	if(flagrandom)
		get_random(&username);

	else if(allname)
		username=allname;

	else if(!pw) {
		if(defname) username=defname;
		else {
			log_error(ci, "could not get user information", 0);
			ident_error(ci,"NO-USER");
			exit(1);
		}
	}

	if(pw) {
		realname=pw->pw_name;
		if(!username) username=realname;
		if(flaguser)
			get_ownname(pw->pw_dir, &username);
	}

	ident_reply(ci,username);
	if(flagverbose) log_success(ci, realname, username);
	exit(0);
}
	
