#include "config.h"
#include "misc.h"
#include "get_uid.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#if HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#if HAVE_SYS_SYSCTL_H
#include <sys/sysctl.h>
#endif

#if HAVE_SYS_UCRED_H
#define _WANT_UCRED
#include <sys/ucred.h>
#endif

/* see, this code requires root privileges. On the other hand, this 
 * information could be retrieved via libkvm, without root privileges.
 * The other approach would make the code 20 times bigger and much more
 * error-prone, not to mention security issues in libkvm. 
 * Why does the net.inet.tcp.getcred sysctl require root privileges is
 * beyond me.
 * --hobbit
 */

int get_uid(struct conn_info c) {

  struct sockaddr_in sin[2];
  struct ucred uc;
  size_t size;

  sin[0].sin_family = sin[1].sin_family = AF_INET;
  sin[0].sin_port = htons(c.lport); 
  memcpy(&sin[0].sin_addr, &c.laddr, sizeof(c.laddr));  
  sin[1].sin_port = htons(c.rport); 
  memcpy(&sin[1].sin_addr, &c.raddr, sizeof(c.raddr)); 

  size = sizeof(uc);

  if (sysctlbyname("net.inet.tcp.getcred", &uc, &size, sin, sizeof(sin)) == -1) 
    return (errno == ENOENT ? H_NO_USER : H_UNKNOWN);
    
  return uc.cr_uid;
    
}
